package Chainmail::MessageType::Rfc822::FieldParser;

use Chainmail::Report;
use Chainmail::Config;
require Chainmail::Address;

my $add_address_to_address_list = 
sub
  {
    my $self = shift;
    my $address = shift;

    enter( "Add address to list hook" );

    if (!defined( $address ))
      {
	error( "sent an undefined address!" );
	leave();
	return;
      }

    
    push @{$self->{ADDRESSES}}, $address;

    leave();
  };

my $call_address_transform =
sub
  {
    my $self = shift;

    my $address = shift;

    return $address->transform();
  };


sub new 
  {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    
    my $self  = {};
    

    @{$self->{IGNORED_TOKENS}} = ( "WHITESPACE" );
    $self->{CURRENT_TOKEN} = undef;

    $self->{TOKENS} = ();
    $self->{TOKEN_STACK} = ();
    $self->{TOKEN_DUMP} = ();

    $self->{CONTEXT} = ();
    $self->{ADDRESSES} = ();
    $self->{ADDRESS_INFO} = undef;

    $self->{ADDRESS_HOOK} = $add_address_to_address_list;

    $self->{DISCARD_TOKENS} = undef;
    
    bless ($self, $class);
    return $self;
  }

sub add_info
  {
    my $self = shift;
    my $new_info = shift;

    if ( defined $self->{ADDRESS_INFO} )
      {
	$self->{ADDRESS_INFO} .= " $new_info";
      }
    else
      {
	$self->{ADDRESS_INFO} = $new_info;
      }
  }

sub lex
  {
    my $self = shift;
    my $text = shift;
    local $_;

    enter( "Lexing RFC 822 header field" );
    debug( "field is:\n\t$text" );

    @{$self->{TOKENS}} = ();
    
    (leave(),return undef)
      if length( $text ) == 0 ;
    
    # Tokens
    
    my $endline = qr/(?:\n|\r\n)/;
    my $linear_ws = qr/(?:$endline)?[ \t]/;
    my $quoted_pair = qr/\\./s;
    my $special_chars = '\(\)\<\>\@\,\;\:\.\[\]\"';
    
    my $atom = qr/[^(?:$special_chars)\t\n ]+/;
    
    my $quoted_string = qr/\"(?:(?:[^\"\\]|$quoted_pair)*)\"/;
    
    my $dtext = qr/[^\r\n\[\]\\\"]/;
    my $domain_literal = qr/\[(?:\$dtext|$quoted_string)*\]/;

    
    
    my %token_types = 
      (
       WHITESPACE => qr/(?:$linear_ws)+/ ,
       QUOTED_STRING => qr/$quoted_string/ ,
       ATOM => qr/$atom/ ,
       DOMAIN_LITERAL => qr/$domain_literal/ ,
       
       AT_SIGN => qr/\@/ ,
       OPEN_ANGLE_BRACKET => qr/\</ ,
       CLOSE_ANGLE_BRACKET => qr/\>/ ,
       OPEN_SQUARE_BRACKET => qr/\[/ ,
       CLOSE_SQUARE_BRACKET => qr/\]/ ,
       
       COLON => ":" ,
       COMMA => ",",
       SEMICOLON => ";",
       STOP => "\\."
      ) ;


    
    $_ = $text;
    chomp;
    
    my $t;
    
  get_token: do 
    {
      
      if ( /^\(/s )
	{
	  ($t,$_) = $self->_strip_leading_comment( $_ );

	  debug( "got comment $t" );
	  push @{$self->{TOKENS}},$self->make_token( "COMMENT",$t ) ;
	}
      
      else
	{
	  foreach $t (keys %token_types)
	    {
	      my $r = $token_types{$t};     
	      
	      if ( /^($r)(.*)/s ) {
		debug( "got token \"$1\" (type $t)" );
		
		push @{$self->{TOKENS}},  $self->make_token( $t, $1 );
		
		$_ = $2;
	      }
	    }  
	}
    } while length;

    leave();

    return 1;
  }

sub make_token
  {

    my $self = shift;
    my $type = shift;
    my $value = shift;

    my $return = { "TYPE"=>$type,"VALUE"=>$value };

    return $return;
  }

sub _strip_leading_comment 
  {
    my $self = shift;
    my $text = shift;

    enter( "Stripping leading comment" );

    my $char;
    my $comment = "";
    
    while ( length $text) 
      {
	$text =~ /^([^\\]|\\.)(.*)/s; # remove first char from string. FIXME   	 
	  $char = $1;
	$text = $2;
	
	$comment .= $char;
	
	$brackets++ 
	  if $char eq "\(" ;
	
	if ( $char eq "\)")
	  {
	    $brackets --;
	    ( leave(),
	    return ( $comment,$text ) )
	      if $brackets == 0 ;
	  }
      }
    error( "unterminated comment (no unquoted closing bracket)" );

    leave();
  }


sub grab_addresses
  {
    my $self = shift;

    enter( "Grabbing addresses from RFC 822 header field" );


    $self->{ADDRESS_HOOK} = $add_address_to_address_list;

    $self->parse( )
      if ( !defined (@{$self->{ADDRESSES}}) );
  
    leave();
  
    return @{$self->{ADDRESSES}}
    unless ( !defined (@{$self->{ADDRESSES}}) );
    
    return undef;
  }


sub transform_addresses
  {
    my $self = shift;

    enter( "Transforming addresses in RFC 822 header field" );

    $self->{ADDRESS_HOOK} = $call_address_transform;

    $self->parse( shift );

    $self->{TOKENS} = $self->{TOKEN_DUMP};
  
    leave();
  
    return 1;
  }


sub parse
  {
    my $self = shift;
    my $text = shift;


    enter( "Parsing RFC 822 address header field" );


    if ( !defined( @{$self->{TOKENS}} ) )
      {
	(leave (), return undef) 
	  if !defined $text;
	$self->lex( $text );
      }

    @{$self->{ADDRESSES}} = ();
    @{$self->{TOKEN_STACK}} = @{$self->{TOKENS}};
    @{$self->{TOKEN_DUMP}} = ();
    $self->{DISCARD_TOKENS} = 0;
    
    (error( "no tokens to read!" ),leave(), return undef) 
      if !defined $self->{TOKEN_STACK};

    $self->{CURRENT_TOKEN} = undef;
    
    $self->get_next_token();
    $self->addresses();

    leave();

    return 1;
  }

sub
get_text
  {
    my $self = shift;
    my $text = undef;

    foreach my $tok (@{$self->{TOKENS}})
      {
	if ( $tok->{TYPE} eq "ADDRESS" )
	  {
	    $text .=  " " . $tok->{VALUE}->get_text() . " ";
	  }
	else
	  {
	    $text .= $tok->{VALUE};
	  }
      }

    return $text;
  }

sub
get_pretty_text
  {
    my $self = shift;
    my $text = undef;

    my $has_ws = 1;

    foreach my $tok (@{$self->{TOKENS}})
      {
	if ( $tok->{TYPE} eq "ADDRESS" )
	  {
	    $text .= " " unless $has_ws;
	    $text .=  $tok->{VALUE}->get_text() . " ";
	    $has_ws = 1;
	  }
	else
	  {
	    if ( $tok->{TYPE} eq "WHITESPACE" )
	      {
		$text .= " " unless $has_ws;
		$has_ws = 1;
	      }
	    elsif ($tok->{TYPE} eq "COMMA" )
	      {
		$text .= $tok->{VALUE} . "\n\t";
	      }
	    else 
	      {
		$text .= $tok->{VALUE};
		$has_ws = 0;
	      }
	  }
      }

    return $text;
  }

sub require_token
  {
    
    
    my $self = shift;
    my $demand = shift;
    
    my $peek_ahead = 11; 
    # Number of tokens to look at to find the $demand.
    
    my $t;

    

    
    
    for (my $i =0; $i < $peek_ahead; $i++ )
      {
	$t = $self->peek_token( $i );
	
	if ( (defined $t) && ($t->{TYPE} eq $demand) )
	  {
	    while ( !(defined ($self->token( $demand )) ) )
	      {
		push @_, $self->{DISCARD_TOKENS};

		$self->{DISCARD_TOKENS} = 1;
		
		$self->parse_error( "discarding token to satisfy demand for $demand" );
		
		$self->get_next_token();

		$self->{DISCARD_TOKENS} = pop @_ ;
	      }

	    return $t->{VALUE};
	    
	  }
      }
    
    $self->parse_error( "could not satisfy request for a token of type $demand even looking ahead $peek_ahead tokens" );

    return undef;
  }

sub request_token
  {
    my $self = shift;
    
    my $request = shift;
    
    return  $self->token( $request );
  }


sub token
  {
    my $self = shift;
    my $request = shift;
    
    defined $self->{CURRENT_TOKEN}
    or return undef;

    $self->skip_ignored_tokens();

    if (( !defined( $request)) || ( $self->{CURRENT_TOKEN}->{TYPE} eq $request  )) 
      {
	my $r = $self->{CURRENT_TOKEN}->{VALUE};
	
	$self->get_next_token();
	return $r;
      }
    
     return undef;
  }

sub skip_ignored_tokens
  {
    my $self = shift;

    if ( $self->peek_token(0)->{TYPE} eq "COMMENT" )
      {
	$self->{CURRENT_TOKEN}->{VALUE} =~ /\(((.)*?)\)$/s
	  || $self->parse_error( "Regexp failed to strip brackets from comment ",  
			   $self->{CURRENT_TOKEN}->{VALUE} );
	
	$self->add_info( $1 )
	  if length $1;
	
	return $self->get_next_token();
      }
    
    foreach my $t (@{$self->{IGNORED_TOKENS}}) 
      {
	if ( $self->{CURRENT_TOKEN}->{TYPE} eq $t )
	  {
	    return $self->get_next_token();
	  }
      }
  }

sub get_next_token
  {
    my $self = shift;
    
    if (defined $self->{CURRENT_TOKEN})
      {
	my $token = shift @{$self->{TOKEN_STACK}};
    
	if ($self->{DISCARD_TOKENS} == 0)
	  {
	    push @{$self->{TOKEN_DUMP}}, $token;
	  }
      }
    

    $self->{CURRENT_TOKEN} = ${$self->{TOKEN_STACK}}[0];

    return undef if !defined $self->{CURRENT_TOKEN};

    debug( "got next token: " . $self->{CURRENT_TOKEN}->{TYPE} );

    
    return $self->{CURRENT_TOKEN};
}

sub token_is
  {
    my $self = shift;

    $self->skip_ignored_tokens();
    
    if ( defined( $self->peek_token (0)) ) {
      return ( $self->peek_token(0)->{TYPE} eq shift );
    }
    else {
      return undef;
    }
    
  }

sub peek_token
  {
    my $self = shift;
    my $index = shift;
    
    return ${$self->{TOKEN_STACK}}[$index];
  }

sub parse_error
  {
    my $self = shift;
    my $message = shift;

    error $message;

    debug (
      "token list is:"
	   );

    foreach $token (@{$self->{TOKEN_STACK}})
      {
	debug(
	  "\t\ttoken type=". $token->{TYPE} .
	  " value=\"".   $token->{VALUE} .
	  "\" address=" . $token 
	     );
      }
    return "";
  }







sub group 
  {
    my $self = shift;

    enter( "Named e-mail group" );
    
    my $group_name = $self->phrase();

    $self->{ADDRESS_INFO} = undef;
    
    $self->require_token( "COLON" );
    
    $self->addresses();
    
    $self->require_token( "SEMICOLON") 
         or $self->parse_error( "named group of addresses apparently not terminated by semi-colon" );

    leave();

  }

sub phrase
  {
    my $self = shift;

    enter( "Phrase; list of words or quoted strings" );

    my $w;
    my $r = undef;

    while ( defined ($w = $self->word()) ) 
      { $r .= "$w " if length $w;}

    chop $r
      if defined $r;

    leave();

    return $r;
  }

sub word
  {
    my $self = shift; 
    return ( $self->request_token( "QUOTED_STRING" ) or $self->request_token( "ATOM" ) or undef);
  }

sub route
  {
    my $self = shift;

    enter( "Parsing route to address" );
    inform( "routed address; route is being ignored" );

    do
      {
	$self->request_token( "AT_SIGN" );
	do
	  {
	    $self->sub_domain();
	  } while( $self->request_token( "STOP" ) );
      }	while( $self->request_token( "COMMA" ) );
	    
    $self->request_token( "COLON" );

    leave();
  }

sub sub_domain
  {
    my $self = shift;
    return ( $self->request_token( "DOMAIN_LITERAL" ) or $self->request_token( "ATOM" ) or undef );
  }

sub route_addr
  {
    my $self = shift;

    enter( "Angle bracketed <address>" );

    $self->request_token( "OPEN_ANGLE_BRACKET" );
    
    if ($self->token_is( "AT_SIGN")) 
      {
	$self->route();
      }
    
    $self->addr_spec();

    $self->require_token( "CLOSE_ANGLE_BRACKET")
      or $self->parse_error( "unterminated < bracket" );

    leave();
  }

sub addr_spec
  {
    my $self = shift;

    enter( "Address (addr-spec\@domain)" );

    my $local_part = ();
    my $domain = ();
    my $t;
    my $addr_obj;

    if ($self->token_is( "ADDRESS" ))
      {
	$addr_obj = $self->token();
	goto addr_obj_got;
      }

    
    do 
      {
	$t = $self->word();

	if ( defined $t )
	  {
	    $t = $1
	      if $t =~ /\"((.)*?)\"$/s;

	    push @{$local_part},  $t;
	  }
	else
	  {
	    $self->parse_error( "not valid address part (token types ATOM or QUOTED_STRING)" );
	  }
      } 
    while ( defined( $self->request_token( "STOP") ) );
    
    $self->request_token ("AT_SIGN")
      or goto address_finished;
    
    do 
      {
	$t = $self->sub_domain();
	if ( !defined $t )
	  {
	    inform( 
	     "missing subdomain part for username @{$local_part} will be filled in" );
	  }
	else
	  {
	    push @{$domain},  $t;
	  }
	    
      } while ( defined ($self->request_token( "STOP") ) );
    
  address_finished:
    $addr_obj = Chainmail::Address->new( $local_part, $domain, $self->{ADDRESS_INFO} );


  addr_obj_got:
    debug( "got address " . $addr_obj->get_text() );

    @_ = ( $self, $addr_obj );

    if ( !defined $self->{ADDRESS_HOOK} )
      {
	error( "address manipulation hook is undefined!" );
      }
    else
      {
	&{$self->{ADDRESS_HOOK}};
      }

    push @{$self->{TOKEN_DUMP}},$self->make_token( "ADDRESS",$addr_obj );

    $self->{ADDRESS_INFO} = undef;

    leave();
  }

sub addresses
  {
    my $self = shift;
    my $state = "Set of comma separated addresses";
    enter( $state );
    do 
      {
	$self->address();
      } while( defined( $self->request_token ("COMMA") ) );
    leave( $state );

  }

sub address
  {
    my $self = shift;
    
    my $token = undef;
    
    for( my $i=0; ( defined( $token = $self->peek_token( $i ) ) ); $i++ )
      {
        
	if ($token->{TYPE} eq "COLON") 
	  {
	    debug( "Guessing group" );
	    
	    return $self->group();
	  }
	
	else 
	  {
	    
	    if ($token->{TYPE} eq "OPEN_ANGLE_BRACKET") 
	      {
		debug( "Guessing <address>" );

		$self->{DISCARD_TOKENS} = 1;

		$self->add_info( $self->phrase() );
		$self->route_addr();

		$self->{DISCARD_TOKENS} = 0;

		return;

	      }
	    else
	      {
		last if ($token->{TYPE} eq "COMMA");
	      }
	  }
      }
    
    debug("Falling back on guess addr_spec\@domain");
    $self->{DISCARD_TOKENS} = 1;
    
    $self->addr_spec();

    $self->{DISCARD_TOKENS} = 0;
    return;
}

1;
